#version 430 core

// Interpolated values from the vertex shaders
in vec2 UV;
in vec3 wsPosition;

in vec3 tsNormal;
in vec3 tsEyeDirection;
in vec3 tsLightDirection;

/////////////////////////////////////////////////////////
in vec4 ShadowCoord;
in vec4 attenuation;
/////////////////////////////////////////////////////////

uniform sampler2D TextureSamplerDiff;
uniform sampler2D TextureSamplerSpec;
uniform sampler2D TextureSamplerNorm;

uniform sampler2D TextureSamplerHeight;

uniform float ambientBrightness;
uniform float normalMappingOnOff;

uniform float shadowThreshold;
uniform sampler2DShadow shadowMap;

uniform int toggleDisplacementOnOff;
uniform	int steps;
uniform	int refinementSteps;

// Ouput data
layout(location = 0) out vec3 color;


vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

float calculateVisibilityShadowMap()
{
	float lit = 1.0f;
	float overdark = 0.75f;
	
	float occluder;
	
	float bias = 0.005; // against shadow acne
	
	for (int i = 0; i < 8; i++) {
		vec3 tmp = vec3(ShadowCoord.xy + poissonDisk[i]/700.0, ShadowCoord.z - bias);
		occluder = texture(shadowMap, tmp);
		lit -= 0.2 * (1.0f - occluder);
	}
	
	lit = clamp(lit, 0.0, 1.0);
	return lit;
}



vec2 getModifiedUV()
{
	vec2 newUV = UV;
	vec2 dUV = normalize(tsEyeDirection).xy * -0.08f; // displacement depth
	
	float prev_hits = 0.0f;
	float hit_h = 0.0f; // the output
	
	float searchHeight = 1.0f;
	
	float correction = steps * 1.0f;
	float refinementCorrection = refinementSteps * 10.0f;
	
	float factor = 1.0f/steps;
	float refinementFactor = 0.1f/refinementSteps;
	
	dUV *= (1.0f/steps); 
	
	// first loop -> coarse steps
	for(int it=0; it < steps; ++it){
		searchHeight -= factor;
		newUV += dUV;
		float h_tex = textureLod(TextureSamplerHeight, newUV, 0).r;
		float is_first_hit = clamp((h_tex - searchHeight - prev_hits)*4999999.0f, 0.0f, 1.0f);
		
		hit_h += is_first_hit * searchHeight;
		prev_hits += is_first_hit;
	}
	
	newUV = UV + dUV * (1.0f - hit_h) * correction - dUV;
	
	vec2 tmpUV = newUV;
	searchHeight = hit_h + 0.1f;
	float start = searchHeight;
	dUV *= (1.0f/refinementSteps);
	prev_hits = 0.0f;
	hit_h = 0.0f;
	
	// second loop -> finer steps
	for(int i=0; i < refinementSteps; ++i){
		searchHeight -= refinementFactor;
		newUV += dUV;
		float h_tex = textureLod(TextureSamplerHeight, newUV, 0).r;
		float is_first_hit = clamp((h_tex - searchHeight - prev_hits)*4999999.0f, 0.0f, 1.0f);
		
		hit_h += is_first_hit * searchHeight;
		prev_hits += is_first_hit;
	}
	
	vec2 modifiedUV = tmpUV + dUV * (start - hit_h) * refinementCorrection;
	
	return modifiedUV;
	
}


/*********************************************************
 * "Global" variable
 *********************************************************/
vec3 diffColor;
vec3 ambientColor;
vec3 specColor;
vec3 textureNormal;

/*********************************************************
 *********************************************************/

 void getColor(vec2 newUV)
 {
	diffColor = texture(TextureSamplerDiff, newUV).rgb;
	ambientColor = diffColor * ambientBrightness;
	specColor = texture(TextureSamplerSpec, newUV).rgb;
    textureNormal = normalize(texture(TextureSamplerNorm, newUV).rgb * 2.0 - 1.0);
 }
 
 
void main()
{
	/************************** HeightMap **************************/
	vec2 tmpUVs = getModifiedUV();
	vec2 newUV = UV;
	if(toggleDisplacementOnOff == 1){
		newUV = tmpUVs;
	}
	/***************************************************************/

    getColor(newUV);
    
    /*********************************************************
	 * Shadow Mapping
	 *********************************************************/
	float vis = calculateVisibilityShadowMap();
    
    /************************* Attenuation *************************/
	float att = 1 - dot(attenuation/10.0f, attenuation/10.0f);
	att = max(att, 0.0f);

	
    /*********************************************************
	 * Normal Mapping
	 *********************************************************/
	float adjustBumpValue = 0.5f;

	vec3 n = (normalMappingOnOff) * normalize(adjustBumpValue * textureNormal + tsNormal) +
		(1.0f - normalMappingOnOff) * normalize(tsNormal);
	
    vec3 l = normalize(tsLightDirection);
	float cosTheta = clamp(dot(n,l), 0, 1);
    float selfshadow = clamp(shadowThreshold * cosTheta, 0, 1);

    /*********************************************************
	 * Specular Mapping
	 *********************************************************/
    vec3 spec = vec3(0.0);
    vec3 eye = normalize(tsEyeDirection);
	float intensity = max(dot(n, l), 0.0);
	
	if(intensity > 0.0) {
        vec3 h = normalize(l + eye);
        float intSpec = max(dot(h,n), 0.0);
        spec = specColor * intSpec; 
    }
	
	vec3 R = reflect(-l,n);
	float cosAlpha = clamp(dot(eye, R), 0, 1);
    
    /*********************************************************
	 * Final Results
	 *********************************************************/
	float visibility = selfshadow * vis * att;

	diffColor = diffColor * cosTheta * visibility;
	specColor = specColor * spec * pow(cosAlpha, 5) * visibility;
    
	color = ambientColor + diffColor + specColor;
}