#version 430 core

in vec3 Position_worldspace;
in vec3 Normal_worldspace;
in vec3 LightDirection_worldspace_point;
in vec3 LightDirection_worldspace_directional;
in vec3 EyeDirection_worldspace;
in vec2 UV;
in vec4 ShadowCoord;
in vec4 attenuation;
in vec4 attenuationD;

uniform vec3 LightPosition_worldspace;
uniform vec2 near_far;
uniform float shadowThreshold;
uniform float ambientBrightness;
uniform float isTorchOn;

uniform sampler2D diffTexture;
uniform sampler2DShadow shadowMap;
uniform samplerCubeShadow shadowCube;

out vec3 color;


vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

/*********************************************************
 * "Global" variable
 *********************************************************/
vec3 ambientColor;
vec3 diffColor;
vec3 specColor;
/*********************************************************
 *********************************************************/

float calculateVisibilityShadowMap()
{
	float lit = 1.0f;
	float overdark = 0.75f;
	
	float occluder;
	
	float bias = 0.005; // against shadow acne
	
	for (int i = 0; i < 8; i++) {
		vec3 tmp = vec3(ShadowCoord.xy + poissonDisk[i]/700.0, ShadowCoord.z - bias);
		occluder = texture(shadowMap, tmp);
		lit -= 0.2 * (1.0f - occluder);
	}
	
	lit = clamp(lit, 0.0, 1.0);
	return lit;
}

float randomAngle(in vec3 seed)
{
	vec2 co = seed.xy;
    return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}


float calculateVisibilityShadowCube()
{
    vec4 lookUpVector = vec4((Position_worldspace) - LightPosition_worldspace, 0);
	float currentDistance = length(lookUpVector) / (near_far.y);
	currentDistance = clamp(currentDistance, 0, 1);
	
	float angle = randomAngle(Position_worldspace);
	float s = sin(angle);
	float c = cos(angle);
	
	float diskRadius = (1.0 + (1.0 - currentDistance) * 3) / (100);
	float compareDistance = currentDistance - 0.005; // bias

	float shadowFactor = 0;
	float numSamples = 0;
    
	for(int i = 0; i < 8; ++i)
	{
		vec2 rotatedOffset = vec2(poissonDisk[i].x * c + poissonDisk[i].y * s, poissonDisk[i].x * -s + poissonDisk[i].y * c);
		vec3 lookup = lookUpVector.xyz + vec3(rotatedOffset * diskRadius, 0);
		lookup = normalize(lookup);
		shadowFactor += texture(shadowCube, vec4(lookup, compareDistance));
		numSamples++;
	}
	
	shadowFactor /= numSamples;
    return clamp(shadowFactor, 0, 1);
}


void getColor()
{
	diffColor = texture(diffTexture, UV).rgb;
	specColor = vec3(0.5, 0.5, 0.5);
	ambientColor = diffColor * ambientBrightness;
}


void getNormalSpecularMapping()
{
	/*********************************************************
	 * Shadow
	 *********************************************************/
	float visD = calculateVisibilityShadowMap();
    float visP = calculateVisibilityShadowCube();

	/*********************************************************
	 * Attenuation
	 *********************************************************/
	float att = 1 - dot(attenuation/10.0f, attenuation/10.0f);
	att = max(att, 0.0f);

	float attD = 1 - dot(attenuationD/10.0f, attenuationD/10.0f);
	attD = max(attD, 0.0f);

	/*********************************************************
	 * Normal Mapping
	 *********************************************************/
	vec3 n = normalize(Normal_worldspace);

	vec3 lP = normalize(LightDirection_worldspace_point);
	vec3 lD = normalize(LightDirection_worldspace_directional);
    
	float cosThetaP = clamp(dot(n, lP), 0, 1);
	float cosThetaD = clamp(dot(n, lD), 0, 1);

    float selfshadowP = clamp(shadowThreshold * cosThetaP, 0, 1);
	float selfshadowD = clamp(shadowThreshold * cosThetaD, 0, 1);

	/*********************************************************
	 * Specular Mapping
	 *********************************************************/
	vec3 eye = normalize(EyeDirection_worldspace);

	float intensityP = max(dot(n, lP), 0.0f);
	float intensityD = max(dot(n, lD), 0.0f);
	vec3 specP = vec3(0.5, 0.5, 0.5);
	vec3 specD = vec3(0.5, 0.5, 0.5);

	if(intensityP > 0.0) {
		vec3 hP = normalize(lP + eye);
		float intSpecP = max(dot(hP, n), 0.0);
		specP = specColor * intSpecP;
	}
	
	if(intensityD > 0.0) {
		vec3 hD = normalize(lD + eye);
		float intSpecD = max(dot(hD, n), 0.0);
		vec3 specD = specColor * intSpecD; 
	}

	vec3 RP = reflect(-lP, n);
    vec3 RD = reflect(-lD, n);
	float cosAlphaP = clamp(dot(eye, RP), 0, 1);
    float cosAlphaD = clamp(dot(eye, RD), 0, 1);

	/*********************************************************
	 * Final Results
	 *********************************************************/
	float visibility = (selfshadowP * visP * 4.0f * (att) * isTorchOn + selfshadowD * 4.0f * visD * attD) * 0.125f;
	float cosTheta =  (cosThetaP * 4.0f + cosThetaD * 4.0f) * 0.125f;
	float cosAlpha =  (cosAlphaP * 4.0f + cosAlphaD * 4.0f) * 0.125f;
	vec3 spec =  (specP * 4.0f + specD * 4.0f) * 0.125f;

	diffColor = diffColor * cosTheta * visibility;
	specColor = specColor * spec * pow(cosAlpha, 5) * visibility;
}

void main()
{
	getColor();
	getNormalSpecularMapping();

	color = ambientColor + diffColor + specColor;
}