#version 430 core

// Interpolated values from the vertex shaders
in mat3 TBN;
in vec4 ShadowCoord;
in vec4 attenuation;
in vec3 Position_worldspace;
in vec3 Normal_worldspace;
in vec3 EyeDirection_worldspace;
in vec3 LightDirection_worldspace;
in vec3 LightDirection_worldspace_directional;
in vec2 UV;

// Ouput data
layout(location = 0) out vec3 color;

// Values that stay constant for the whole mesh.
uniform sampler2D DiffuseTextureSampler;
uniform sampler2D NormalTextureSampler;
uniform sampler2D SpecularTextureSampler;

uniform samplerCubeShadow shadow;
uniform sampler2DShadow shadowMap;

uniform vec3 LightPosition_worldspace;
uniform vec3 adjustBumpValue;
uniform vec2 near_far;
uniform int BumpOn;
uniform float shadowThreshold;

vec2 poissonDisk[16] = vec2[]( 
   vec2( -0.94201624, -0.39906216 ), 
   vec2( 0.94558609, -0.76890725 ), 
   vec2( -0.094184101, -0.92938870 ), 
   vec2( 0.34495938, 0.29387760 ), 
   vec2( -0.91588581, 0.45771432 ), 
   vec2( -0.81544232, -0.87912464 ), 
   vec2( -0.38277543, 0.27676845 ), 
   vec2( 0.97484398, 0.75648379 ), 
   vec2( 0.44323325, -0.97511554 ), 
   vec2( 0.53742981, -0.47373420 ), 
   vec2( -0.26496911, -0.41893023 ), 
   vec2( 0.79197514, 0.19090188 ), 
   vec2( -0.24188840, 0.99706507 ), 
   vec2( -0.81409955, 0.91437590 ), 
   vec2( 0.19984126, 0.78641367 ), 
   vec2( 0.14383161, -0.14100790 ) 
);

float randomAngle(in vec3 seed)
{
	vec2 co = seed.xy;
    return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

float calculateVisibilityShadowMap()
{
	float lit = 1.0f;
	float overdark = 0.75f;
	
	float occluder;
	float receiver;
	
	float bias = 0.005; // against shadow acne
	vec4 tmp;
	
	for (int i = 0; i < 8; i++)
	{
		tmp = vec4(ShadowCoord.xy + poissonDisk[i]/700.0, ShadowCoord.z - bias, ShadowCoord.w);
		occluder = texture(shadowMap, tmp.xyz);
		receiver = (ShadowCoord.z)/ShadowCoord.w;
		lit -= 0.2 * (1.0f - exp(overdark * occluder) * exp(-1.0f * overdark * receiver));
	}
	
	lit = clamp(lit, 0.0, 1.0);
	return lit;
}


float calculateVisibilityShadowCube()
{
    vec4 lookUpVector = vec4((Position_worldspace) - LightPosition_worldspace, 0);
	float currentDistance = length(lookUpVector) / (near_far.y);
	currentDistance = clamp(currentDistance, 0, 1);
	
	float angle = randomAngle(Position_worldspace);
	float s = sin(angle);
	float c = cos(angle);
	
	float diskRadius = (1.0 + (1.0 - currentDistance) * 3) / (100);
	float compareDistance = currentDistance - 0.005; // bias

	float shadowFactor = 0;
	float numSamples = 0;
    
	for(int i = 0; i < 8; ++i)
	{
		vec2 rotatedOffset = vec2(poissonDisk[i].x * c + poissonDisk[i].y * s, poissonDisk[i].x * -s + poissonDisk[i].y * c);
		vec3 lookup = lookUpVector.xyz + vec3(rotatedOffset * diskRadius, 0);
		lookup = normalize(lookup);
		shadowFactor += texture(shadow, vec4(lookup, compareDistance));
		numSamples++;
	}
	
	shadowFactor /= numSamples;
    return clamp(shadowFactor, 0, 1);
}



void main()
{
    vec3 LightColor = vec3(1,1,1);
	float LightPower = 1.0f;
	
	
	/**************************** Shadow ***************************/
	float visP = calculateVisibilityShadowCube();
    float visD = calculateVisibilityShadowMap();
	
	/************************* Attenuation *************************/
	float att = 1 - dot(attenuation/10.0f, attenuation/10.0f); // square version
	// must not be negative, 0 is already no light!
	att = max(att, 0.0f);

	/************************* Materials ***************************/
	vec3 MaterialDiffuseColor = texture(DiffuseTextureSampler, UV).rgb;
	vec3 MaterialAmbientColor = vec3(0.1,0.1,0.1) * MaterialDiffuseColor;
	vec3 MaterialSpecularColor = texture( SpecularTextureSampler, UV).rgb * 0.9;
	
    /************************* Bump mapping ***********************/
	vec3 TextureNormal_tangentspace = normalize(texture( NormalTextureSampler, vec2(UV.x, UV.y) ).rgb * 2.0 - 1.0);
	vec3 n = (BumpOn) * normalize( transpose(TBN) * (adjustBumpValue * TextureNormal_tangentspace) + Normal_worldspace) +
		(1-BumpOn) * normalize(Normal_worldspace);
    
    vec3 lD = normalize(LightDirection_worldspace_directional);
	vec3 lP = normalize(LightDirection_worldspace);
    
    float cosThetaD = clamp(dot(n, lD), 0, 1);
	float cosThetaP = clamp(dot(n, lP), 0, 1);
	float selfshadowD = clamp(shadowThreshold * cosThetaD, 0, 1);
    float selfshadowP = clamp(shadowThreshold * cosThetaP, 0, 1);

	/************************* Specular ***************************/

	vec3 eye = normalize(EyeDirection_worldspace);

	float intensityP = max(dot(n, lP), 0.0f);
	float intensityD = max(dot(n, lD), 0.0f);
	vec3 specP = vec3(0,0,0);
	vec3 specD = vec3(0,0,0);

	if(intensityP > 0.0) {
		vec3 hP = normalize(lP + eye);
		float intSpecP = max(dot(hP, n), 0.0);
		specP = MaterialSpecularColor * intSpecP;
	}
	
	if(intensityD > 0.0) {
		vec3 hD = normalize(lD + eye);
		float intSpecD = max(dot(hD, n), 0.0);
		vec3 specD = MaterialSpecularColor * intSpecD; 
	}
    

	vec3 RP = reflect(-lP, n);
    vec3 RD = reflect(-lD, n);
	float cosAlphaP = clamp(dot(eye, RP), 0, 1);
    float cosAlphaD = clamp(dot(eye, RD), 0, 1);

    /*********************** Final colors *************************/
    vec3 DiffuseD = MaterialDiffuseColor * cosThetaD;
    vec3 SpecularD = MaterialSpecularColor * specD * pow(cosAlphaD,5);
    vec3 ShadowD = selfshadowD * visD * LightColor * LightPower;
    vec3 colorD = MaterialAmbientColor + ShadowD * (DiffuseD + SpecularD);

    vec3 Diffuse = MaterialDiffuseColor * cosThetaP;
    vec3 Specular = MaterialSpecularColor * specP * pow(cosAlphaP, 5);
	vec3 Shadow = selfshadowP * visP * LightColor * LightPower;
    vec3 colorP = MaterialAmbientColor + Shadow * (Diffuse + Specular) * (att);

	color = (colorP * 5.0f + colorD * 3.0f) * 0.125f;
}